import torch
from torch import nn
from torch.nn import functional as F
import warnings

# noinspection PyUnreachableCode
if False:
    from typing import Any, Callable, Tuple, Union

from . import backend


class Signature(nn.Module):
    """Module wrapper around the signatory.signature function. See :func:`signatory.signature`.

    Arguments:
        depth: as :func:`signatory.signature`.

        basepoint: as :func:`signatory.signature`.

        stream: as :func:`signatory.signature`.

    Called with a single argument :attr:`path` of type :class:`torch.Tensor`.
    """

    def __init__(self, depth, basepoint=False, stream=False, **kwargs):
        # type: (int, bool, bool, **Any) -> None
        if not isinstance(depth, int) or depth < 1:
            raise ValueError('Depth must be an integer greater than or equal to one. Given {depth} of type '
                             '{tdepth}'.format(depth=depth, tdepth=type(depth)))
        super(Signature, self).__init__(**kwargs)
        self.depth = depth
        self.basepoint = basepoint
        self.stream = stream

    def forward(self, path):
        # type: (torch.Tensor) -> Union[torch.Tensor, Tuple[torch.Tensor, ...]]
        if path.size(1) == 1:
            warnings.warn('{clsname} called on path with only one channel, so the signature is now just the moments of '
                          'the path.'.format(clsname=self.__class__.__name__))
        return backend.signature(path, self.depth, self.stream, self.basepoint)

    def extra_repr(self):
        return 'depth={depth}, stream={stream}, basepoint={basepoint}'.format(depth=self.depth, stream=self.stream,
                                                                              basepoint=str(self.basepoint)[:6])


class Augment(nn.Module):
    r"""Augmenting the stream before feeding it into a signature is often useful; the hope is to obtain higher-order
    information in the signature. One way to do this is in a data-dependent way is to apply a feedforward neural network
    to sections of the stream, so as to obtain another stream; on this stream the signature is then applied.

    The input path is expected to be a three-dimensional tensor, with dimensions :math:`(N, L, C)`, where :math:`N` is
    the batch size, :math:`L` is the length of the input sequence, and :math:`C` denotes the number of channels. Thus
    each batch element is interpreted as a stream of data :math:`(x_1, \ldots, x_L)`, where each
    :math:`x_i \in \mathbb{R}^C`. (This is the same as :class:`torch.nn.Conv1d`, for example.)

    Then this stream may be 'augmented' via some function

    .. math::
        \phi \colon \mathbb{R}^{C \times k} \to \mathbb{R}^{\widehat{C}}

    giving a stream of data

    .. math::
        \left(\phi(x_1, ... x_k), \ldots, \phi(x_{n - k + 1}, \ldots, x_n)\right),

    which is essentially a three-dimensional tensor with dimensions :math:`(N, L - k + 1, \widehat{C})`.

    Thus this essentially operates as a one dimensional convolution, except that a whole network is swept across the
    input, rather than just a single convolutional layer.

    Both the original stream and time can be specifically included in the augmentation. (This usually tends to give
    better empirical results.) For example, if both :attr:`include_original` is True and :attr:`include_time` is True,
    then each :math:`\phi(x_i, ... x_{k + i - 1})` is of the form

    .. math::
        \left(\frac{i}{T}, x_i, \varphi(x_i, ... x_{k + i - 1})\right).

    where :math:`T` is a constant appropriately chosen so that the first entry moves between :math:`0` and :math:`1` as
    :math:`i` varies. (Specifically, :math:`T = L - k + 1 + 2 \times \text{padding}`.)

    For further details see `Deep Signatures -- Bonnier et al. 2019 <https://arxiv.org/abs/1905.08494>`_. (This module
    is here more for convenience: it doesn't directly relate to the signature transform; it's just useful to have around
    when you are using the signature transform.)

    Arguments:
        in_channels (int): Number of channels :math:`C` in the input stream.

        layer_sizes (tuple of int): Specifies the sizes of the layers of the feedforward neural network to apply to
            the stream. The final value of this tuple specifies the number of channels in the augmented stream,
            corresponding to the value :math:`\widehat{C}` in the preceding discussion.

        kernel_size (int): Specifies the size of the kernel to slide over the stream, corresponding to the value
            :math:`k` in the preceding discussion.

        stride (int, optional): Defaults to 1. How far to move along the input stream before re-applying the
            feedforward neural network. Thus the output stream is given by

            .. math::
                (\phi(x_1, \ldots, x_k),
                 \phi(x_{1 + \text{stride}}, \ldots, x_{k + 2 \times \text{stride}}),
                 \phi(x_{1 + 2 \times \text{stride}}, \ldots, x_{k + 2 \times \text{stride}}),
                 \ldots)

        padding (int, optional): Defaults to 0. How much zero padding to add to either end of the the input stream
            before sweeping the feedforward neural network along it.

        dilation (int, optional): The spacing between input elements given to the feedforward neural network.
            Defaults to 1. Harder to describe; see the equivalent argument for :class:`torch.nn.Conv1d`.

        bias (bool, optional): Defaults to True. Whether to use biases in the neural network.

        activation (callable, optional): Defaults to ReLU. The activation function to use in the feedforward neural
            network.

        include_original (bool, optional): Defaults to True. Whether or not to include the original stream
            (pre-augmentation) in the augmented stream.

        include_time (bool, optional): Defaults to True. Whether or not to also augment the stream with a 'time' value.
            These are values in :math:`[0, 1]` corresponding to how far along the stream dimension the element is.

    .. note::

        Thus the resulting stream of data has shape :math:`(N, L, \text{out_channels})`, where in pseudocode:

        .. code-block:: python

            out_channels = layer_sizes[-1]
            if include_original:
                out_channels += in_channels
            if include_time:
                out_channels += 1

    """

    def __init__(self,
                 in_channels,            # type: int
                 layer_sizes,            # type: Tuple[int, ...]
                 kernel_size,            # type: int
                 stride=1,               # type: int
                 padding=0,              # type: int
                 dilation=1,             # type: int
                 bias=True,              # type: bool
                 activation=F.relu,      # type: Callable[[torch.Tensor], torch.Tensor]
                 include_original=True,  # type: bool
                 include_time=True,      # type: bool
                 **kwargs                # type: Any
                 ):
        # type: (...) -> None
        super(Augment, self).__init__(**kwargs)

        if isinstance(layer_sizes, int):
            layer_sizes = (layer_sizes,)

        self.in_channels = in_channels
        self.layer_sizes = layer_sizes
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.bias = bias
        self.activation = activation
        self.include_original = include_original
        self.include_time = include_time

        self.convs = nn.ModuleList()
        if layer_sizes:
            self.convs.append(nn.Conv1d(in_channels=in_channels,
                                        out_channels=layer_sizes[0],
                                        kernel_size=kernel_size,
                                        stride=stride,
                                        padding=padding,
                                        dilation=dilation,
                                        bias=bias))
            last_layer_channels = layer_sizes[0]
            for augment_channel in layer_sizes[1:]:
                # These pointwise convolutions correspond to sliding a standard feedforward network across the input.
                self.convs.append(nn.Conv1d(in_channels=last_layer_channels,
                                            out_channels=augment_channel,
                                            kernel_size=1,
                                            bias=bias))
                last_layer_channels = augment_channel

    def forward(self, x):
        # type: (torch.Tensor) -> torch.Tensor
        if len(x.shape) != 3:
            raise RuntimeError('Argument x should have three dimensions, (batch, stream, channel). Given shape'
                               '{shape} dimensions with {x}.'.format(shape=x.shape, x=x))
        len_truncated = x.size(1) - self.kernel_size + 1
        pieces = []
        if self.include_original:
            truncated_x = x.narrow(1, self.kernel_size - 1, len_truncated)
            pieces.append(truncated_x)

        if self.include_time:
            time = torch.linspace(0, 1, len_truncated, dtype=torch.float, device=x.device)
            time.unsqueeze_(1)
            time = time.expand(x.size(0), len_truncated, 1)
            pieces.append(time)

        if self.layer_sizes:
            augmented_x = self.convs[0](x.transpose(1, 2))
            for conv in self.convs[1:]:
                augmented_x = self.activation(augmented_x)
                augmented_x = conv(augmented_x)
            augmented_x.transpose_(1, 2)
            pieces.append(augmented_x)
        return torch.cat(pieces, dim=2)  # concatenate along channel axis

    def extra_repr(self):
        return ('activation={activation}, include_original={include_original}, include_time={include_time}'
                ).format(activation=self.activation, include_original=self.include_original,
                         include_time=self.include_time)



import torch
from torch import autograd

# noinspection PyUnreachableCode
if False:
    from typing import Any, Tuple, Union

from ._impl import (_signature_channels,
                    _signature_forward,
                    _signature_backward)


# It would be lovely to do all of this at the C++ level. (In particular sigspec is really a struct that has no
# business being passed around at the Python level.) But unfortunately the documentation for how to create autograd
# Functions in C++ is nonexistent. Presumably that means it's all still subject to change, so we're just going to stick
# to the Python way of doings things for now.
class _SignatureFunction(autograd.Function):
    @staticmethod
    def forward(ctx, path, depth, stream, basepoint, basepoint_value):
        # type: (Any, torch.Tensor, int, bool, torch.Tensor, bool) -> Union[torch.Tensor, Tuple[torch.Tensor, ...]]
        result, backwards_info = _signature_forward(path, depth, stream, basepoint, basepoint_value)
        ctx._backwards_info = backwards_info
        ctx._basepoint = basepoint
        return result

    @staticmethod
    @autograd.function.once_differentiable  # Our backward function uses in-place operations for memory efficiency
    def backward(ctx, grad_result):
        # type: (Any, Tuple[torch.Tensor]) -> Tuple[torch.Tensor, None, None, None, Union[None, torch.Tensor]]
        grad_path, grad_basepoint_value = _signature_backward(grad_result, ctx._backwards_info)
        return grad_path, None, None, None, grad_basepoint_value


def signature(path, depth, stream=False, basepoint=False):
    # type: (torch.Tensor, int, bool, Union[bool, torch.Tensor]) -> Union[torch.Tensor, Tuple[torch.Tensor, ...]]
    r"""Applies the signature transform to a stream of data.

    The input :attr:`path` is expected to be a three-dimensional tensor, with dimensions :math:`(N, L, C)`, where
    :math:`N` is the batch size, :math:`L` is the length of the input sequence, and :math:`C` denotes the number of
    channels. Thus each batch element is interpreted as a stream of data :math:`(x_1, \ldots, x_L)`, where each
    :math:`x_i \in \mathbb{R}^C`. (This is the same as :class:`torch.nn.Conv1d`, for example.)

    If :attr:`basepoint` is True then an additional point :math:`x_0 = 0 \in \mathbb{R}^C` is prepended to the path.
    (Alternatively it can be a :class:`torch.Tensor1` of shape :math:`(N, C)` specifying the point to prepend.)

    Each path is then lifted to a piecewise linear path :math:`X \colon [0, 1] \to \mathbb{R}^C`, and the signature
    transform of this path is then computed. This (by the definition of the signature transform) gives a sequence of
    tensors of shape

    .. math::
        (N, C), (N, C, C), \ldots (N, C, \ldots, C),

    where the final tensor has :attr:`depth` many dimensions of size :math:`C`. These are then flattened down and
    concatenated to give a single tensor of shape

    .. math::
        (N, C + C^2 + \cdots + C^\text{depth}).

    (This value may be computed via the :func:`signatory.signature_channels` function.)

    If :attr:`stream` is True then  the signatures of all paths :math:`(x_1, \ldots, x_j)`, for :math:`j=2, \ldots, L`,
    are computed. (Or :math:`(x_0, \ldots, x_j)`, for :math:`j=1, \ldots, L` if :attr:`basepoint` is provided. In
    neither case is the signature of the path of a single element computed, as that isn't defined.)

    Examples:
        If :attr:`stream` is False then the returned tensor will have shape

        .. math::
            (N, C + C^2 + \cdots + C^d).

        If :attr:`basepoint` is True and :attr:`stream` is True then the returned tensor will
        have shape

        .. math::
            (N, L, C + C^2 + \cdots + C^d),

        as the stream dimension is now preserved. See also the 'Returns' section below.

    Arguments:
        path (:class:`torch.Tensor`): The batch of input paths to apply the signature transform to.

        depth (int): The depth to truncate the signature at.

        stream (bool, optional): Defaults to False. If False then the signature transform of the whole path is computed.
            If True then the signature of all intermediate paths are also computed.

        basepoint (bool or :class:`torch.Tensor`, optional): Defaults to False. If True, then the input paths will have
            an additional point at the origin prepended to the start of the sequence. (If this is False then the
            signature transform is invariant to translations of the path, which may or may not be desirable.)
            Alternatively it may be a :class:`torch.Tensor` specifying the point to prepend, in which case it should
            have shape :math:`(N, C)`

    Returns:
        A :class:`torch.Tensor`. Given an input :class:`torch.Tensor` of shape :math:`(N, L, C)`, and input arguments
        :attr:`depth`, :attr:`basepoint`, :attr:`stream`, then the return value is, in pseudocode:

        .. code-block:: python

            if stream:
                if basepoint is True or isinstance(basepoint, torch.Tensor):
                    return torch.Tensor of shape (N, L, C + C^2 + ... + C^depth)
                else:
                    return torch.Tensor of shape (N, L - 1, C + C^2 + ... + C^depth)
            else:
                return torch.Tensor of shape (N, C + C^2 + ... + C^depth)

    """

    if basepoint is True:
        basepoint_value = torch.zeros((path.shape[0], path.shape[2]), dtype=path.dtype, device=path.device)
    elif isinstance(basepoint, torch.Tensor):
        basepoint_value = basepoint
        basepoint = True
    else:
        basepoint_value = torch.Tensor()
    return _SignatureFunction.apply(path, depth, stream, basepoint, basepoint_value)


# A wrapper for the sake of consistent documentation on signatures
def signature_channels(in_channels, depth):
    # type: (int, int) -> int
    """Computes the number of output channels from a signature call.

    Arguments:
        in_channels (int): The number of channels in the input; that is, the dimension of the space that the input path
            resides in.

        depth (int): The depth of the signature that is being computed.

    Returns:
        An int specifying the number of channels in the signature of the path.
    """

    return _signature_channels(in_channels, depth)


def extract_term(signature, in_channels, depth):
    # type: (torch.Tensor, int, int) -> torch.Tensor
    r"""Extracts a particular term from a signature.

    The signature to depth :math:`d` of a batch of paths in :math:`\mathbb{R}^\text{C}` is a tensor with
    :math:`C + C^2 + \cdots + C^d` channels. (See :func:`signatory.signature`.) This function extracts the :attr:`depth`
    term of that, returning a tensor with just :math:`C^\text{depth}` channels.

    This is really just here as a convenience function: if you want to extract multiple terms and care about speed then
    it will be probably quicker to write your own function.

    Arguments:
        signature (:class:`torch.Tensor`): The signature to extract the term from. Should be the result of the
            :func:`signatory.signature` function.

        in_channels (int): The number of input channels :math:`C`.

        depth (int): The depth of the term to be extracted from the signature.

    Returns:
        The :class:`torch.Tensor` corresponding to the :attr:`depth` term of the signature.
    """

    if depth == 1:
        start = 0
    else:
        start = signature_channels(in_channels, depth - 1)
    return signature.narrow(dim=-1, start=start, length=in_channels ** depth)



import random
import signatory
import torch
import torch.autograd as autograd
import unittest


class TestGrad(unittest.TestCase):
    @staticmethod
    def gradcheck(size=(1, 4, 3), depth=2, stream=False, basepoint=False):
        path = torch.rand(*size, requires_grad=True, dtype=torch.double)
        return autograd.gradcheck(signatory.signature, (path, depth, stream, basepoint))

    def test_gradcheck_edge(self):
        for stream in (True, False):
            for size in ((1, 2, 1), (1, 4, 4), (4, 2, 4), (4, 4, 1)):
                basepoint_size = (size[0], size[2])
                for basepoint in (True,
                                  False,
                                  torch.rand(basepoint_size, requires_grad=True, dtype=torch.double),
                                  torch.rand(basepoint_size, requires_grad=False, dtype=torch.double)):
                    for depth in (1, 2):
                        try:
                            self.gradcheck(size, depth, stream, basepoint)
                        except RuntimeError:
                            self.fail("Failed with stream={stream}, basepoint={basepoint}, size={size}, depth={depth}, "
                                      "basepoint_size={basepoint_size}"
                                      .format(stream=stream, basepoint=basepoint, size=size, depth=depth,
                                              basepoint_size=basepoint_size))

    def test_gradcheck_random(self):
        for stream in (True, False):
            for _ in range(5):
                for basepoint_ in (True, False, None):
                    size = torch.randint(low=1, high=10, size=(3,))
                    size[1] += 1  # stream dimension must be at least size 2
                    basepoint_size = (size[0], size[2])
                    if basepoint_ is None:
                        basepoint_grad = random.choice([True, False])
                        basepoint = torch.rand(basepoint_size, requires_grad=basepoint_grad, dtype=torch.double)
                    else:
                        basepoint_grad = None
                        basepoint = basepoint_
                    depth = int(torch.randint(low=1, high=4, size=(1,)))
                    try:
                        self.gradcheck(size, depth, stream, basepoint)
                    except Exception:
                        self.fail("Failed with stream={stream}, basepoint={basepoint}, size={size}, depth={depth}, "
                                  "basepoint_size={basepoint_size}, basepoint_grad={basepoint_grad}"
                                  .format(stream=stream, basepoint=basepoint, size=size, depth=depth,
                                          basepoint_size=basepoint_size, basepoint_grad=basepoint_grad))

    # We don't do gradgradcheck because our backwards function uses a whole bunch of in-place operations for memory
    # efficiency, so it's not automatically differentiable. (And I'm not writing a custom double backward function...)



import iisignature
import signatory
import torch
import unittest


class TestAccuracy(unittest.TestCase):
    def test_forward(self):
        for stream in (True, False):
            for N in (1, 2, 3, 4):
                for L in (2, 3, 4, 5):
                    for C in (1, 2, 3, 4):
                        for basepoint in (True, False, torch.rand(N, C, dtype=torch.float64)):
                            size = (N, L, C)
                            path = torch.rand(size, requires_grad=True, dtype=torch.float64)
                            depth = int(torch.randint(low=1, high=4, size=(1,)))
                            out = signatory.signature(path, depth, stream=stream, basepoint=basepoint)
                            if basepoint is True:
                                ii_in = torch.cat([torch.zeros(N, 1, C, dtype=torch.float64), path.detach()], dim=1)
                            elif basepoint is False:
                                ii_in = path.detach()
                            else:
                                ii_in = torch.cat([basepoint.unsqueeze(1), path.detach()], dim=1)
                            iiout = iisignature.sig(ii_in, depth, 2 if stream else 0)
                            close = out.allclose(torch.tensor(iiout, dtype=torch.float64))
                            if not close:
                                self.fail("out={out}, iiout={iiout}".format(out=out, iiout=iiout))

    def test_backward(self):
        for N in (1, 2, 3, 4):
            for L in (2, 3, 4, 5):
                for C in (1, 2, 3, 4):
                    for basepoint in (True,
                                      False,
                                      torch.rand(N, C, requires_grad=True, dtype=torch.float64),
                                      torch.rand(N, C, requires_grad=False, dtype=torch.float64)):
                        size = (N, L, C)
                        path = torch.rand(size, requires_grad=True, dtype=torch.float64)
                        depth = int(torch.randint(low=1, high=4, size=(1,)))
                        out = signatory.signature(path, depth, stream=False, basepoint=basepoint)
                        grad = torch.rand_like(out)
                        out.backward(grad)
                        gradresult = path.grad
                        if basepoint is True:
                            ii_in = torch.cat([torch.zeros(N, 1, C, dtype=torch.float64), path], dim=1)
                        elif basepoint is False:
                            ii_in = path
                        else:  # implies isinstance(basepoint, torch.Tensor) == True
                            ii_in = torch.cat([basepoint.unsqueeze(1), path], dim=1)
                            if basepoint.requires_grad:
                                # get the gradient on the basepoint as well
                                gradresult = torch.cat([basepoint.grad.unsqueeze(1), gradresult], dim=1)
                        iisig_backward = iisignature.sigbackprop(grad, ii_in.detach(), depth)
                        if basepoint is True or (isinstance(basepoint, torch.Tensor) and not basepoint.requires_grad):
                            # if we don't have a gradient through the basepoint then discard the corresponding basepoint
                            # part of the iisig_backward result
                            iisig_backward = iisig_backward[:, 1:, :]
                        # strangely iisignature returns float32 in the backward calculation, even if the input was
                        # float64, so we have to reduce the tolerance slightly
                        close = gradresult.allclose(torch.tensor(iisig_backward, dtype=torch.float64), atol=1e-6)
                        if not close:
                            self.fail("gradresult={gradresult}, iisig_backward={iisig_backward}, basepoint={basepoint} "
                                      .format(gradresult=gradresult, iisig_backward=iisig_backward,
                                              basepoint=basepoint))



import signatory
import torch
import unittest


class TestArguments(unittest.TestCase):
    def test_minimal_axes(self):
        depth = 2
        for stream in (True, False):
            for size in ((1, 4, 4), (4, 2, 4), (4, 4, 1), (1, 2, 1)):
                basepoint_size = (size[0], size[2])
                for basepoint in (True,
                                  False,
                                  torch.rand(basepoint_size, requires_grad=True),
                                  torch.rand(basepoint_size, requires_grad=False)):
                    try:
                        out = signatory.signature(torch.rand(size, requires_grad=True), depth, stream, basepoint)
                        out.backward(torch.rand_like(out))
                    except Exception:
                        self.fail("stream={stream}, basepoint={basepoint}, size={size}, depth={depth}, "
                                  "basepoint_size={basepoint_size}"
                                  .format(stream=stream, basepoint=basepoint, size=size, depth=depth,
                                          basepoint_size=basepoint_size))

    def test_too_small_axes(self):
        depth = 2
        for stream in (True, False):
            for size in ((0, 4, 4), (4, 1, 4), (4, 0, 4), (4, 4, 0), (1, 1, 1)):
                basepoint_size = (size[0], size[2])
                for basepoint in (True,
                                  False,
                                  torch.rand(basepoint_size, requires_grad=True),
                                  torch.rand(basepoint_size, requires_grad=False)):
                    with self.assertRaises(ValueError):
                        signatory.signature(torch.rand(size, requires_grad=True), depth, stream, basepoint)

    def test_arguments(self):
        for stream in (True, False):
            for N in (1, 2, 3, 4):
                for L in (1, 2, 3, 4):
                    for C in (1, 2, 3, 4):
                        size = (N, L, C)
                        basepoint_size = (N, C)
                        for basepoint in (True,
                                          False,
                                          torch.rand(basepoint_size, requires_grad=True),
                                          torch.rand(basepoint_size, requires_grad=False)):
                            path = torch.rand(size, requires_grad=True)
                            depth = int(torch.randint(low=1, high=4, size=(1,)))
                            try:
                                if L == 1:
                                    with self.assertRaises(ValueError):
                                        signatory.signature(path, depth, stream, basepoint)
                                else:
                                    out = signatory.signature(path, depth, stream, basepoint)
                                    out.backward(torch.rand_like(out))
                            except Exception:
                                self.fail("stream={stream}, basepoint={basepoint}, size={size}, depth={depth}, "
                                          "basepoint_size={basepoint_size}"
                                          .format(stream=stream, basepoint=basepoint, size=size, depth=depth,
                                                  basepoint_size=basepoint_size))


class TestShapes(unittest.TestCase):
    @staticmethod
    def correct_shape(size, depth, stream, basepoint):
        N, L, C = size
        if stream:
            if isinstance(basepoint, torch.Tensor) or basepoint:
                return N, L, signatory.signature_channels(C, depth)
            else:
                return N, L - 1, signatory.signature_channels(C, depth)
        else:
            return N, signatory.signature_channels(C, depth)

    def test_shapes(self):
        for stream in (True, False):
            for N in (1, 2, 3, 4):
                for L in (2, 3, 4, 5):
                    for C in (1, 2, 3, 4):
                        basepoint_size = (N, C)
                        for basepoint in (True,
                                          False,
                                          torch.rand(basepoint_size, requires_grad=True),
                                          torch.rand(basepoint_size, requires_grad=False)):
                            size = (N, L, C)
                            path = torch.rand(size)
                            depth = int(torch.randint(low=1, high=4, size=(1,)))
                            out = signatory.signature(path, depth, stream, basepoint)
                            correct_shape = self.correct_shape(size, depth, stream, basepoint)
                            self.assertEqual(out.shape, correct_shape, "stream={stream}, basepoint={basepoint}, "
                                                                       "N={N}, L={L}, C={C}, depth={depth},"
                                                                       "basepoint_size={basepoint_size}"
                                                                       .format(stream=stream, basepoint=basepoint,
                                                                               N=N, L=L, C=C, depth=depth,
                                                                               basepoint_size=basepoint_size))



import signatory
import iisignature
import timeit
import torch
import unittest
import warnings


class TestSpeed(unittest.TestCase):
    @staticmethod
    def speed(batch_size=1000, number=10, depth=3, stream=True, backward=False):
        if stream and backward:
            raise ValueError("iisignature does not support backpropagation for a full stream of data, so we can't "
                             "compare against them.")
        signatory_x = torch.rand(batch_size, 100, 10, requires_grad=backward)
        iisignature_x = signatory_x.detach().numpy()

        def signatory_fn():
            y = signatory.signature(signatory_x, depth, stream=stream)
            if backward:
                y.backward(torch.rand_like(y))

        def iisignature_fn():
            for iisignature_ix in iisignature_x:
                y = iisignature.sig(iisignature_ix, depth, 2 if stream else 0)
                if backward:
                    iisignature.sigbackprop(torch.rand_like(torch.tensor(y)), iisignature_ix, depth)

        signatory_time = timeit.timeit(signatory_fn, number=number)
        iisignature_time = timeit.timeit(iisignature_fn, number=number)
        return signatory_time, iisignature_time, signatory_time / iisignature_time

    def wrapped_speed_test(self, stream, backward, **kwargs):
        signatory_time, iisignature_time, ratio = self.speed(stream=stream, backward=backward, **kwargs)
        self.assertLess(ratio, 1.0)
        # can usually expect closer to 0.7 if backward else 0.4, but this gives a bit of safety margin
        if ratio > (0.8 if backward else 0.6):
            warnings.warn("Speed was unusually slow: signatory time: {stime} iisignature time: {itime}, ratio: {ratio} "
                          "stream: {stream}, backward: {backward}".format(stime=signatory_time, itime=iisignature_time,
                                                                          ratio=ratio, stream=stream,
                                                                          backward=backward))

    def test_speed_forward_nostream(self):
        self.wrapped_speed_test(number=20, stream=False, backward=False)

    def test_speed_forward_stream(self):
        self.wrapped_speed_test(number=20, stream=True, backward=False)

    def test_speed_backward(self):
        self.wrapped_speed_test(number=20, stream=False, backward=True)
